/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */


package org.experlog.gencode;

import java.util.Stack;

import java.io.PushbackReader;
import java.io.StringReader;
import java.io.IOException;

public class ExpParser {

  String theString_ = null;
  PushbackReader str_;
  double currentValue_;
  String currentStringValue_;
  int currentValueType_ = 0; // 0 for Numeric, 1 for String
  int previousToken_ = -1;

  public boolean evalexpr_ = true; // Evaluate boolean expressions

  public static final int EOL     = 0;
  public static final int VALUE   = 1;
  public static final int OPAR  = 2;
  public static final int CPAR  = 3;
  public static final int EXP     = 4;
  public static final int MULT    = 5;
  public static final int DIV     = 6;
  public static final int PLUS    = 7;
  public static final int MINUS   = 8;
  public static final int AND   = 9;
  public static final int OR   = 10;
  public static final int NOT   = 11;
  public static final int EQUAL   = 12;
  public static final int DIFF   = 13;
  public static final int GT   = 14;
  public static final int LT   = 15;
  public static final int GE   = 16;
  public static final int LE   = 17;
  public static final int UMINUS   = 18;

  Stack stack_ = new Stack();
  String warning_ = null;

  /**
   * Create a new ExpParser
   * Parse both boolean and numeric expressions
   * @param s The expression to parse
   */
  public ExpParser(String s) {
    this(s, true);
  }

  /**
   * Create a new ExpParser
   * Parse both boolean and numeric expressions, or numeric expressions only
   * @param s The expression to parse
   * @param evalexpr If false, only numeric expressions are parsed
   * (boolean expressions are then considered as strings)
   */
  public ExpParser(String s, boolean evalexpr) {
    evalexpr_ = evalexpr;
    theString_ = s;
    if(s == null || s.trim().length() < 1) { 
      str_ = null;
      theString_ = null;
    }
    else str_ = new PushbackReader(new StringReader(s), 2);
  }

  public void setWarning(String w) { warning_ = w; }
  /**
   * Get and reset warning
   */
  public String getWarning() {
    String w = null;
    if(warning_ != null) {
      w = new String(warning_);
      warning_ = null;
    }
    return w;
  }

  public void setEvalExpr(boolean eval) {
    evalexpr_ = eval;
  }

  public String getRawString() { return theString_; }

  public ExpOperand buildFormula(int minsize) throws IOException {

    if(str_ == null) return null;

    int preccrt = -1;
    int tok = -1;

    do {
      tok = getToken();

      //System.out.println("Stack=" + stack_ + ", new tok=" + tok);
      //System.out.println("Prec: crt=" + preccrt
      // + ", new tok=" + getPrecedence(tok));

      if(tok == EOL || tok == CPAR) {}
      else if(tok == OPAR) {
        stack_.push(buildFormula(stack_.size()+1));
      }

      else if(tok != VALUE && getPrecedence(tok) <= preccrt) {

        if(! isUnary(tok)) { // Unary ops will just be stacked
          try {
            ExpOperand opright = toOperand(stack_.pop());
            int operator = toOperator(stack_.pop());
            ExpOperand opleft = null;
            if(!isUnary(operator)) opleft = toOperand(stack_.pop());
            stack_.push(new ExpOperand(operator, opleft, opright)); // Push result
          } catch(Exception e) {
            stack_.empty();
            throw new IOException("Parse error during evaluation");
          }
        }

        stack_.push(new Integer(tok)); // Push operator
        preccrt = getPrecedence(tok);

      } else {
        if(tok == VALUE) {
          if(currentValueType_ == 0) stack_.push(new Double(currentValue_));
          else stack_.push(currentStringValue_);
        } else {
          stack_.push(new Integer(tok)); // Push operator
          preccrt = getPrecedence(tok);
        }
      }

    } while(tok != EOL && tok != CPAR);

    while(stack_.size() > minsize) {
      try {
        ExpOperand opright = toOperand(stack_.pop());
        int operator = toOperator(stack_.pop());
        ExpOperand opleft = null;
        if(!isUnary(operator)) opleft = toOperand(stack_.pop());
        stack_.push(new ExpOperand(operator, opleft, opright)); // Push result
      } catch(Exception e) {
        stack_.empty();
        //e.printStackTrace();
        throw new IOException("Parse error during evaluation");
      }
    }

    //System.out.println("Final Stack=" + stack_ + ",tok=" + tok);
    try {
      return toOperand(stack_.pop());
    } catch(Exception e) {
      stack_.empty();
      throw new IOException("Parse error during evaluation");
    }
  }

  public boolean isUnary(int operator) {
    return (operator == NOT || operator == UMINUS);
  }

  public int getToken() throws IOException {

    int c = str_.read();
    while(c == ' ' || c == '\t') c = str_.read();

    if(c < 0) {
      str_.close();
      return EOL;
    }

    int tok;

    // Strings are enclosed in simple quotes
    // Exception: multiple operator chars (like ++ or */) will be considered
    // strings, see extractString() method below...
    if(c == '\'') {
      StringBuffer sb = new StringBuffer();
      boolean go = true;
      while(go && c >= 0) {
        c = str_.read();
        if(c == '\'') {
          int next = str_.read();
          go = (next == '\'');
          if(!go) { if(next >= 0) str_.unread(next); }
          else sb.append('\'');
        } else {
          sb.append((char)c);
        }
      }
      if(go && c < 0) {
        str_.close();
        return EOL;
      }

      currentStringValue_ = sb.toString();
      currentValueType_ = 1;
      tok = VALUE;
    }

    else if(c == '^') tok = (extractString("^") ? VALUE : EXP);
    else if(c == '/') tok = (extractString("/") ? VALUE : DIV);
    else if(c == '*') tok = (extractString("*") ? VALUE : MULT);
    else if(c == '(') tok = OPAR;
    else if(c == ')') tok = CPAR;
    else if(c == '+') tok = (extractString("+") ? VALUE : PLUS);
    else if(c == '-') tok = (extractString("-") ? VALUE : (previousToken_ == VALUE || previousToken_ == CPAR ? MINUS : UMINUS));
    else if(evalexpr_ && c == '#') tok = DIFF;
    else if(evalexpr_ && c == '&') {
      int nextc = str_.read();
      if(nextc == '&') tok = AND;
      else {
        if(nextc >= 0) str_.unread(nextc);
        tok = AND;
      }
    }
    else if(evalexpr_ && c == '|') {
      int nextc = str_.read();
      if(nextc == '|') tok = OR;
      else {
        if(nextc >= 0) str_.unread(nextc);
        tok = OR;
      }
    }
    else if(evalexpr_ && c == '=') {
      int nextc = str_.read();
      if(nextc == '=') tok = EQUAL;
      else {
        if(nextc >= 0) str_.unread(nextc);
        tok = EQUAL;
      }
    }
    else if(evalexpr_ && c == '!') {
      int nextc = str_.read();
      if(nextc == '=') tok = DIFF;
      else {
        if(nextc >= 0) str_.unread(nextc);
        tok = NOT;
      }
    }
    else if(evalexpr_ && c == '>') {
      int nextc = str_.read();
      if(nextc == '=') tok = GE;
      else {
        if(nextc >= 0) str_.unread(nextc);
        tok = GT;
      }
    }
    else if(evalexpr_ && c == '<') {
      int nextc = str_.read();
      if(nextc == '=') tok = LE;
      else {
        if(nextc >= 0) str_.unread(nextc);
        tok = LT;
      }
    }
    else {

      StringBuffer sb = new StringBuffer();
      sb.append((char)c);
      while(!isOperatorChar(c = str_.read())) sb.append((char)c);
      // TBD
      if(sb.length() >= 2 && sb.charAt(sb.length()-1) == 'E' && Character.isDigit(sb.charAt(sb.length()-2))) {
        sb.append((char)c);
        while(!isOperatorChar(c = str_.read())) sb.append((char)c);
      }
      if(c >= 0) str_.unread(c); // Push back operator char

      String s = sb.toString().trim();
      try {
        currentValue_ = Double.parseDouble(s);
        currentValueType_ = 0;
      } catch(NumberFormatException e) {
        int lastpos = s.length()-1;
        if(s.charAt(0)=='\'' && s.charAt(lastpos)=='\'')
          s = s.substring(1, lastpos);
        currentStringValue_ = s;
        currentValueType_ = 1;
      }
      tok = VALUE;
    }

    previousToken_ = tok; // to detect unary -
    return tok;
  }

  /**
   * Multiple numeric operator chars will be considered a string
   * For example, ***  or +-- will be equivalent to '***' or '+--'
   * Beware, some constructs like +- or *- are legal in numeric expressions
   * (like in 3+-2), so these are not considered strings.
   * A string that starts with 2 or more operator chars will also be
   * valid, like --test (but any space char will end it).
   * This is only for backward compatibility...
   * Note that values like a*b will cause parse errors (must be enclosed
   * in simple quotes, like in 'a*b')
   * Also, comparison operators are excluded, as well as parentheses.
   */
  private boolean extractString(String prefix) throws IOException {
    int c = str_.read();

    if(c < 0) {
      currentStringValue_ = prefix;
      currentValueType_ = 1;
      setWarning("Dubious string without quotes in expression [" + theString_ + "]");
      return true;
    }

    if("+*-/^".indexOf((char)c) < 0) {
      str_.unread(c);
      return false;
    }

    // 2 operators + a digit are legal in numeric expressions: eg. 2+-3
    int cd = str_.read();
    if(Character.isDigit((char)cd)) {
      str_.unread(cd);
      str_.unread(c);
      return false;
    }

    StringBuffer sb = new StringBuffer(prefix);
    sb.append((char)c);

    // Stop if operator other than +*-/^, or if space char
    while(("+*-/^".indexOf((int)(c = str_.read())) >= 0 || (c != ' ' && c != '\t' && ! isOperatorChar(c))) && c >= 0)
      sb.append((char)c);
    if(c >= 0) str_.unread(c);

    currentStringValue_ = sb.toString().trim();
    currentValueType_ = 1;
    setWarning("Dubious string without quotes in expression [" + theString_ + "]");
    return true;
  }

  boolean isOperatorChar(int c) {
    if(c == -1) return true; // End of stream
    if(evalexpr_)
      return ("+*-/^()&|!=><#".indexOf((char)c) >= 0);
    else
      return ("+*-/^()".indexOf((char)c) >= 0);
  }

  public int toOperator(Object o) {
    return ((Integer)o).intValue();
  }

  public ExpOperand toOperand(Object o) {
    if(o instanceof Double) return new ExpOperand((Double)o);
    else if(o instanceof String) return new ExpOperand((String)o);
    else return (ExpOperand)o;
  }

  public int getPrecedence(int token) {
    switch(token) {
      case OR:
        return 1;
      case AND:
        return 2;
      case EQUAL:
      case DIFF:
      case GT:
      case LT:
      case GE:
      case LE:
        return 10;
      case PLUS:
      case MINUS:
        return 11;
      case MULT:
      case DIV:
        return 12;
      case EXP:
        return 13;
      case UMINUS:
        return 14;
      case NOT:
        return 15;
    }
    return 0;
  }

  public static String opString(int op) {
    switch(op) {
      case EXP: return "^";
      case MULT: return "*";
      case DIV: return "/";
      case PLUS: return "+";
      case MINUS: return "-";
      case AND: return "&&";
      case OR: return "||";
      case NOT: return "!";
      case EQUAL: return "==";
      case DIFF: return "#";
      case GT: return ">";
      case LT: return "<";
      case GE: return ">=";
      case LE: return "<=";
      case UMINUS: return "-";
    }
    return "?";
  }

  public ExpOperand parse() throws IOException {
    return buildFormula(1);
  }


  public static void main(String args[]) throws Exception {

    java.io.BufferedReader reader;
    if(args.length > 0)
      reader = new java.io.BufferedReader(new java.io.FileReader(args[0]));
    else
      reader = new java.io.BufferedReader(
                    new java.io.InputStreamReader(System.in));

    String formula;
    while((formula = reader.readLine()) != null) {

      System.out.print("INPUT: " + formula);
      ExpParser parser = new ExpParser(formula, false);
      ExpOperand result = parser.parse();

      System.out.print(" PARSE TREE: "
        + (result == null ? "null" : result.toString()));
      ExpEval eval = new ExpEval();
      System.out.println(" EVAL: " + eval.eval(result));
    }
  }

}

